<?php

namespace Drupal\jsonapi_extras;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\jsonapi\IncludeResolver;
use Drupal\jsonapi\JsonApiResource\JsonApiDocumentTopLevel;
use Drupal\jsonapi\JsonApiResource\LinkCollection;
use Drupal\jsonapi\JsonApiResource\NullEntityCollection;
use Drupal\jsonapi\JsonApiResource\ResourceObject;
use Drupal\jsonapi\ResourceType\ResourceTypeRepositoryInterface;
use Drupal\jsonapi\Serializer\Serializer;
use Symfony\Component\Serializer\SerializerInterface;

/**
 * Simplifies the process of generating a JSON:API version of an entity.
 *
 * @api
 */
class EntityToJsonApi {

  /**
   * The currently logged in user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * Serializer object.
   *
   * @var \Drupal\jsonapi\Serializer\Serializer
   */
  protected $serializer;

  /**
   * The JSON:API resource type repository.
   *
   * @var \Drupal\jsonapi\ResourceType\ResourceTypeRepositoryInterface
   */
  protected $resourceTypeRepository;

  /**
   * The JSON:API include resolver.
   *
   * @var \Drupal\jsonapi\IncludeResolver
   */
  protected $includeResolver;

  /**
   * EntityToJsonApi constructor.
   *
   * @param \Drupal\jsonapi\Serializer\Serializer|\Drupal\jsonapi_extras\SerializerDecorator $serializer
   *   The serializer.
   * @param \Drupal\jsonapi\ResourceType\ResourceTypeRepositoryInterface $resource_type_repository
   *   The resource type repository.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The currently logged in user.
   * @param \Drupal\jsonapi\IncludeResolver $include_resolver
   *   The JSON:API include resolver.
   */
  public function __construct(SerializerInterface $serializer, ResourceTypeRepositoryInterface $resource_type_repository, AccountInterface $current_user, IncludeResolver $include_resolver) {
    assert($serializer instanceof Serializer || $serializer instanceof SerializerDecorator);
    $this->serializer = $serializer;
    $this->resourceTypeRepository = $resource_type_repository;
    $this->currentUser = $current_user;
    $this->includeResolver = $include_resolver;
  }

  /**
   * Return the requested entity as a raw string.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to generate the JSON from.
   * @param string[] $includes
   *   The list of includes.
   *
   * @return string
   *   The raw JSON string of the requested resource.
   */
  public function serialize(EntityInterface $entity, array $includes = []) {
    return $this->serializer->encode($this->normalize($entity, $includes), 'api_json');
  }

  /**
   * Return the requested entity as an structured array.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to generate the JSON from.
   * @param string[] $includes
   *   The list of includes.
   *
   * @return array
   *   The JSON structure of the requested resource.
   */
  public function normalize(EntityInterface $entity, array $includes = []) {
    $resource_type = $this->resourceTypeRepository->get($entity->getEntityTypeId(), $entity->bundle());
    $resource_object = new ResourceObject($resource_type, $entity);
    $context = $this->calculateContext($resource_object);
    $document = $this->prepareDocument($resource_object, $includes);
    return $this->serializer->normalize($document,
      'api_json',
      $context
    )->getNormalization();
  }

  /**
   * Calculate the arguments for the serialize/normalize operation.
   *
   * @param \Drupal\jsonapi\JsonApiResource\ResourceObject $resource_object
   *   The resource object to generate the JSON from.
   *
   * @return array
   *   The list of arguments for serialize/normalize operation.
   */
  protected function calculateContext(ResourceObject $resource_object) {
    return [
      'account' => $this->currentUser,
      'resource_object' => $resource_object,
    ];
  }

  /**
   * Prepare the JsonApi Document for serialization or normalization.
   *
   * @param \Drupal\jsonapi\JsonApiResource\ResourceObject $resource_object
   *   The resource object to generate the JSON from.
   * @param array $includes
   *   An array of included related entities to add to the document.
   *
   * @return \Drupal\jsonapi\JsonApiResource\JsonApiDocumentTopLevel
   *   The top level document.
   */
  private function prepareDocument(ResourceObject $resource_object, array $includes) {
    $includes = !empty($includes)
      ? $this->includeResolver->resolve($resource_object, implode(',', $includes))
      : new NullEntityCollection();
    return new JsonApiDocumentTopLevel($resource_object, $includes, new LinkCollection([]));
  }

}
